# This program was made as part of the sk1project activity to improve UniConvertor
# See http://www.sk1project.org
#
# The code of vsview based on the results of the ALT Linux vsdump project.
# See www.altlinux.ru and http://freshmeat.net/projects/vsdump
#
# Copyright (C) 2007,	Valek Filippov (frob@df.ru)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 or later of the GNU General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import vsdoc
import struct
import tree
import math


## Signature: (doc, pntr, page)
# 46 -- MM, 30 -- "DT" (Inches), 30 -- PT, 49 --??? what's this "IN_F", 50 -- ??? what's this "DA", maybe it's RADs
# 51 -- DEG, degree?

def Geometry(doc,data,page):
    offFlags = 19
    flags = ord(data[offFlags])
    if flags&1 == 1:
        page.shapes[page.curshape].geom.nofill = 1
    if flags&2 == 1:
        page.shapes[page.curshape].geom.noline = 1
    if flags&4 == 1:
        page.shapes[page.curshape].geom.noshow = 1
    if flags&8 == 1:
        page.shapes[page.curshape].geom.nosnap = 1

def Fill(doc,data,page):
    offFG = 19
    offBG = 25
    offFPat = 29 # byte
    offShFG = 31
    offShBG = 35 # 3 bytes
    offShPat = 40 # byte
    offShOffX = 43 # ieee
    offShOffY = 52 # ieee
    offShScale = 69 # ieee
    FG = ord(data[offFG])
    r = ord(data[offBG])
    g = ord(data[offBG+1])
    b = ord(data[offBG+2])
    fgcolor = doc.colors[FG]
    bgcolor = vsdoc.color()
    bgcolor.r = r
    bgcolor.g = g
    bgcolor.b = b
    page.shapes[page.curshape].fill.set = 1
    page.shapes[page.curshape].fill.fgcolor = fgcolor
    page.shapes[page.curshape].fill.fgcolor.a = 1
    page.shapes[page.curshape].fill.bgcolor = bgcolor
##    print 'FG color num: %x, color: #%02x%02x%02x\
##      BG color: #%02x%02x%02x'%(FG,fgcolor.r,fgcolor.g,fgcolor.b,r,g,b)

def Line(doc,data,page):
    offFlags = 50
    offClr = 28
    offWidth = 19
    offPattern = 33
    flags = ord(data[offFlags])
    clr = ord(data[offClr])
    color = doc.colors[clr]
    wunit = ord(data[offWidth])
    pattern = ord(data[offPattern])
##    coeff = wunits[wunit]
    [width] = struct.unpack('<d',data[offWidth+1:offWidth+9])
    page.shapes[page.curshape].line.lineclr = color
    page.shapes[page.curshape].line.flags = flags
    page.shapes[page.curshape].line.width = width*25.4
    page.shapes[page.curshape].line.pattern = pattern
##    print 'Flags: %x. Line color num: %x, color: #%02x%02x%02x'%(flags,clr,color.r,color.g,color.b)
##    print 'Line width: ', 25.4*width,'mm.  Pattern: %x.  Units: %x'%(pattern,wunit)
    
def Char(doc,data,page):
##0 30 Bold
##1 30 Italic
##2 30 Underline
##3 30 Smallcaps
##    [LangID] = struct.unpack('<h', data[offLangID:offLangID+2])
##   fbold = 0
##   fslant = 0
##   funder = 0
    
    offClr = 25
    offSize = 37
    offFont = 23
    offLangID = 88
    offFlags = 30
    clr = ord(data[offClr])
    color = doc.colors[clr]
    [size] = struct.unpack('<d', data[offSize:offSize+8])
    [Font] = struct.unpack('<h', data[offFont:offFont+2])
    flags = ord(data[offFlags])
    str = ''
    if flags&1 == 1:
##        page.shapes[page.curshape].char.fbold = 1
        str = 'Bold'
    if flags&2 == 2:
##        page.shapes[page.curshape].char.fslant = 1
        str+=' Italic'
    if flags&4 == 4:
        page.shapes[page.curshape].char.funder = 1
##        str+=' Underlined'
    page.shapes[page.curshape].char.style = str
    page.shapes[page.curshape].char.charclr = color
    page.shapes[page.curshape].char.fntsize = size*25.4 ##typo point to mm
    page.shapes[page.curshape].char.fntnum = Font
    if str == '':   
        str = 'Regular'
##    print 'FontSize: %.3f  %x'%(size,ord(data[36]))
##    print 'FontNum: %x. Modifications: %s'%(Font, str)
##    print 'Char color num: %x, color: #%02x%02x%02x'%(clr,color.r,color.g,color.b)
    
    
def Text(doc,data,page):
    offset = 27
    if doc.version == 11:
        str=unicode(data[offset:],'utf-16').encode('utf-8')
    else:
        str = unicode(data[offset:],'cp1251').encode('utf-8')
    page.shapes[page.curshape].text.txt = str
##    print 'Text: ',str

def Para(doc,data,page):
    offInd1st = 24 # 8
    offIndLeft = 33 # 8
    offIndRight = 42 # 8
    offSpLine = 51 # 8
    offSpBefore = 59 # 8
    offSpAfter = 69 # 8
    offHAlign = 77 # 1
    offBullet = 78 # 1
##    offBulletStr =  # Need to parse blocks for that
    para = vsdoc.Para()
    [Ind1st] = struct.unpack('<d', data[offInd1st:offInd1st+8])
    [IndLeft] = struct.unpack('<d', data[offIndLeft:offIndLeft+8])
    [IndRight] = struct.unpack('<d', data[offIndRight:offIndRight+8])
    [SpLine] = struct.unpack('<d', data[offSpLine:offSpLine+8])
    [SpBefore] = struct.unpack('<d', data[offSpBefore:offSpBefore+8])
    [SpAfter] = struct.unpack('<d', data[offSpAfter:offSpAfter+8])
    HAlign = ord(data[offHAlign])
    Bullet = ord(data[offBullet])
    para.ind1st = Ind1st
    para.indLeft = IndLeft
    para.indRight = IndRight
    para.spLine = SpLine
    para.spBefore = SpBefore
    para.spAfter = SpAfter
    para.HAlign = HAlign
    para.Bullet = Bullet
    page.shapes[page.curshape].text.para = para
##    print 'Para. Ind 1st/L/R: %u %u %u. Space L/B/A: %u %u %u. HAlign: %u. Bullet: %u'%(Ind1st,IndLeft,IndRight,SpLine,SpBefore,SpAfter,HAlign,Bullet)

def TextBlock(doc,data,page):
    offLeftM = 20 # 8
    offRightM = 29 # 8
    offTopM = 38 # 8
    offBtmM = 47 # 8
    offVAlign = 55 # 1
    offTxtBG = 56 # 3 bytes color
    offDfltTab = 62 # 8
    offTxtDir = 82 # 1
    [LeftM] = struct.unpack('<d', data[offLeftM:offLeftM+8])
    [RightM] = struct.unpack('<d', data[offRightM:offRightM+8])
    [TopM] = struct.unpack('<d', data[offTopM:offTopM+8])
    [BottomM] = struct.unpack('<d', data[offBtmM:offBtmM+8])
    VAlign = ord(data[offVAlign])
    TxtBG = vsdoc.color()
    TxtBG.r = ord(data[offTxtBG])
    TxtBG.g = ord(data[offTxtBG+1])
    TxtBG.b = ord(data[offTxtBG+2])
    [DfltTab] = struct.unpack('<d', data[offDfltTab:offDfltTab+8])
    TxtDir = ord(data[offTxtDir])
    txtblock = vsdoc.TxtBlock()
    txtblock.LMarg = LeftM
    txtblock.RMarg = RightM
    txtblock.TMarg = TopM
    txtblock.BMarg = BottomM
    txtblock.VAlign = VAlign
    txtblock.TxtBG = TxtBG
    txtblock.DfltTab = DfltTab
    txtblock.TxtDir = TxtDir
    page.shapes[page.curshape].text.txtblock = txtblock
##    print 'TxtBlock. Margins L/R/T/B: %u %u %u %u. VAlign: %u. TxtBG: %02x%02x%02x. DfltTab: %u. TxtDir: %u'%(LeftM,RightM,TopM,BottomM,VAlign,TxtBG.r,TxtBG.g,TxtBG.b,DfltTab,TxtDir)
    
def MoveTo(doc,data,page):
    offX = 20
    offY = 29
    [x] = struct.unpack('<d', data[offX:offX+8])
    [y] = struct.unpack('<d', data[offY:offY+8])
##    cr.move_to(x*25.4,-y*25.4)
    cmd = vsdoc.cmd()
    cmd.type = 'MoveTo'
    cmd.args = (25.4*x,25.4*y)
    page.shapes[page.curshape].geom.cmds.append(cmd)
##    print "MoveTo: %u %u"%(x*25.4,y*25.4)

def ArcTo(doc,data,page):
    offX = 20
    offY = 29
    offA = 38
    neg = 0
    [x] = struct.unpack('<d', data[offX:offX+8])
    [y] = struct.unpack('<d', data[offY:offY+8])
    [a] = struct.unpack('<d', data[offA:offA+8])
    x = 25.4*x
    y = 25.4*y
    a = 25.4*a
    cmd = vsdoc.cmd()
    cmd.type = 'ArcTo'
    cmd.args = (x,y,a)
    page.shapes[page.curshape].geom.cmds.append(cmd)
##    print "ArcTo: %u %u %u"%(x,y,180*a/math.pi)

def EllipticalArcTo(doc,data,page):
    offX = 20
    offY = 29
    offA = 38
    offB = 47
    offC = 56
    offD = 65
    [x] = struct.unpack('<d', data[offX:offX+8])
    [y] = struct.unpack('<d', data[offY:offY+8])
    [a] = struct.unpack('<d', data[offA:offA+8])
    [b] = struct.unpack('<d', data[offB:offB+8])
    [c] = struct.unpack('<d', data[offC:offC+8])
    [d] = struct.unpack('<d', data[offD:offD+8])
    x = 25.4*x
    y = 25.4*y
    a = 25.4*a
    b = 25.4*b
    cmd = vsdoc.cmd()
    cmd.type = 'EllipticalArcTo'
    cmd.args = (x,y,a,b,c,d)
    page.shapes[page.curshape].geom.cmds.append(cmd)
##    print "EllipticalArcTo: %u %u %u %u %u %u"%(x,y,a,b,c,d)
    
def LineTo(doc,data,page):
    offset1 = 20
    offset2 = 29
    [x] = struct.unpack('<d', data[offset1:offset1+8])
    [y] = struct.unpack('<d', data[offset2:offset2+8])
    cmd = vsdoc.cmd()
    cmd.type = 'LineTo'
    cmd.args = (25.4*x,25.4*y)
    page.shapes[page.curshape].geom.cmds.append(cmd)
##    print "LineTo: %u %u"%(x*25.4,y*25.4)

def PolylineTo(doc,data,page):
    offset1 = 20
    offset2 = 29
    [x] = struct.unpack('<d', data[offset1:offset1+8])
    [y] = struct.unpack('<d', data[offset2:offset2+8])
    cmd = vsdoc.cmd()
    cmd.type = 'PolylineTo'
    cmd.args = (25.4*x,25.4*y)
    page.shapes[page.curshape].geom.cmds.append(cmd)
##    print "PolylineTo: %u %u"%(x*25.4,y*25.4)

def Ellipse(doc,data,page):
    offX = 20
    offY = 29
    offA = 38
    offB = 47
    offC = 56
    offD = 65
    [x] = struct.unpack('<d', data[offX:offX+8])
    [y] = struct.unpack('<d', data[offY:offY+8])
    [a] = struct.unpack('<d', data[offA:offA+8])
    [b] = struct.unpack('<d', data[offB:offB+8])
    [c] = struct.unpack('<d', data[offC:offC+8])
    [d] = struct.unpack('<d', data[offD:offD+8])
    cmd = vsdoc.cmd()
    cmd.type = 'Ellipse'
    cmd.args = (25.4*x,25.4*y,25.4*a,25.4*b,25.4*c,25.4*d)
    page.shapes[page.curshape].geom.cmds.append(cmd)
##    print 'Ellipse: %u %u %u %u %u %u'%(x*25.4,y*25.4,a*25.4,b*25.4,c*25.4,d*25.4)

def XForm1d(doc,data,page):
    offBX = 20
    offBY = 29
    offEX = 38
    offEY = 47
    [bx] = struct.unpack('<d', data[offBX:offBX+8])
    [by] = struct.unpack('<d', data[offBY:offBY+8])
    [ex] = struct.unpack('<d', data[offEX:offEX+8])
    [ey] = struct.unpack('<d', data[offEY:offEY+8])
    xform1d = vsdoc.XForm1D()
    xform1d.bx =25.4*bx
    xform1d.by =25.4*by
    xform1d.ex =25.4*ex
    xform1d.ey =25.4*ey
    page.shapes[page.curshape].xform1d = xform1d
    
def XForm(doc,data,page):
    offX = 20 # PinX
    offY = 29 # PinY
    offW = 38
    offH = 47
    offLX = 56 # LocPinX
    offLY = 65 # LocPinY
    offAng = 74
    offFlipX = 82
    offFlipY = 83
    
    [x] = struct.unpack('<d', data[offX:offX+8])
    [y] = struct.unpack('<d', data[offY:offY+8])
    [w] = struct.unpack('<d', data[offW:offW+8])
    [h] = struct.unpack('<d', data[offH:offH+8])
    [lx] = struct.unpack('<d', data[offLX:offLX+8])
    [ly] = struct.unpack('<d', data[offLY:offLY+8])
    fx = ord(data[offFlipX])
    fy = ord(data[offFlipY])
    [ang] = struct.unpack('<d', data[offAng:offAng+8])
    xform = vsdoc.XForm()
    xform.x =25.4*x
    xform.y =25.4*y
    xform.w =25.4*w
    xform.h =25.4*h
    xform.locx =25.4*lx
    xform.locy =25.4*ly
    xform.flipx =fx
    xform.flipy =fy
    xform.ang = ang
    page.shapes[page.curshape].xform = xform
##    print 'XForm X/Y/W/H/LX/LY/Ang/fx/fy: %u %u %u %u %u %u %.2f %u %u'%(x*25.4,y*25.4,w*25.4,h*25.4,lx*25.4,ly*25.4,180*ang/math.pi,fx,fy)

def TxtXForm(doc,data,page):
    offX = 20 # PinX
    offY = 29 # PinY
    offW = 38
    offH = 47
    offLX = 56 # LocPinX
    offLY = 65 # LocPinY
    offAng = 74
    [x] = struct.unpack('<d', data[offX:offX+8])
    [y] = struct.unpack('<d', data[offY:offY+8])
    [w] = struct.unpack('<d', data[offW:offW+8])
    [h] = struct.unpack('<d', data[offH:offH+8])
    [lx] = struct.unpack('<d', data[offLX:offLX+8])
    [ly] = struct.unpack('<d', data[offLY:offLY+8])
    [ang] = struct.unpack('<d', data[offAng:offAng+8])
    if page.curshape != -1:
        txtxform = vsdoc.TxtXForm()
        txtxform.x =25.4*x
        txtxform.y =25.4*y
        txtxform.w =25.4*w
        txtxform.h =25.4*h
        txtxform.locx =25.4*lx
        txtxform.locy =25.4*ly
        txtxform.ang = ang
        page.shapes[page.curshape].text.txtxform = txtxform
##    print 'TxtXForm X/Y/Ang: %u %u %.2f'%((x-w/2)*25.4,(y-h/2)*25.4,180*ang/math.pi)


def PageSheet(doc,data,page):
    shape = vsdoc.shape()
    [ix] = struct.unpack('<L',data[4:8])
    if ix == 0xFFFFFFFF:
        ix = len(page.shapes)
    shape.type = 0x46
    page.shapes[ix] = shape
    page.curshape=ix
##    page.shapes[page.curshape].type  = 0x46 ## Page
##    print '----------------------SHAPE!-[%u]------------------------'%page.curshape
    

def Shape(doc,data,page):
    shape = vsdoc.shape()
    [stidx] = struct.unpack('<L',data[0x25:0x29])
    shape.Foreign.Stencil = stidx
    [ix] = struct.unpack('<L',data[4:8])
    if ix == 0xFFFFFFFF:
        ix = len(page.shapes)
    shape.IX = ix
    page.shapes[ix] = shape
    page.curshape=ix
    page.shapes[page.curshape].type  = 0x48 ## Shape
##    print '----------------------SHAPE!-[%u]------------------------'%page.curshape
    
def PageProps(doc,data,page):
    offW = 20
    offH = 29
    [w] = struct.unpack('<d', data[offW:offW+8])
    [h] = struct.unpack('<d', data[offH:offH+8])
    page.width = w*25.4
    page.height = h*25.4
##    print 'PageProps H/W: %u %u'%(h*25.4,w*25.4)

def ShapeID(doc,data,page): ##0x83 -- take ShapeID and append it to Page.ShapeList
    offID = 19
##    offType = 23
    [ID] = struct.unpack('<L', data[offID:offID+4])
    page.shapes[page.curshape].ShapeList.append(ID)
##    print 'Append ShapeID: ',ID

def Group(doc,data,page):
    shape = vsdoc.shape()
    [stidx] = struct.unpack('<L',data[0x25:0x29])
    shape.Foreign.Stencil = stidx
    [ix] = struct.unpack('<L',data[4:8])
    if ix == 0xFFFFFFFF:
        ix = len(page.shapes)
    shape.IX = ix
    page.shapes[ix] = shape
    page.curshape=ix
    page.shapes[page.curshape].type  = 0x47 ## Group
##    print '----------------------SHAPE GROUP!-[%u]------------------'%page.curshape

def Foreign(doc,data,page):
    shape = vsdoc.shape()
    [stidx] = struct.unpack('<L',data[0x25:0x29])
    shape.Foreign.Stencil = stidx
    [ix] = struct.unpack('<L',data[4:8])
    if ix == 0xFFFFFFFF:
        ix = len(page.shapes)
    shape.IX = ix
    page.shapes[ix] = shape
    page.curshape=ix
    page.shapes[page.curshape].type  = 0x4e ## Foreign
    foreign = vsdoc.Foreign()
    page.shapes[page.curshape].foreign  = foreign
##    print '----------------------SHAPE FOREIGN!-[%u]----------------'%page.curshape

def ForeignData(doc,data,page):
    [size] = struct.unpack('<L', data[12:16])
##    doc.stmap[page.shapes[page.curshape].Foreign.Sig] = len(doc.stencils)
##    print 'Stencil mapping: %x %x'%(page.shapes[page.curshape].Foreign.sig,len(doc.stencils))
    page.shapes[page.curshape].Foreign.Data = data[19:19+size]
    cmd = vsdoc.cmd()
    cmd.type = 'Foreign'
    page.shapes[page.curshape].geom.cmds.append(cmd)
##    print 'Foreign Data extracted.'

def Icon(doc,data,page):
## port vsd_parse_stream23.c here
    palette = {}
    size = struct.pack('<I',ord(data[15+6])*ord(data[15+8])*5/8+0x68)
    h = ord(data[15+6])*2
##    print 'Icon: %x %x'%(ord(data[15+6]),ord(data[15+8]))
    icodata = '\x00\x00\x01\x00\x01\x00'+data[15+8]+data[15+6]+'\x10\x00\x01\x00\x04\x00'+size[0:4]+\
    '\x16\x00\x00\x00\x28\x00\x00\x00'+data[15+8]+'\x00\x00\x00'+'\x40'+'\x00\x00\x00'+\
    '\x01\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'+\
    '\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x80\x00\x00'+\
    '\x00\x80\x80\x00\x80\x00\x00\x00\x80\x00\x80\x00\x80\x80\x00\x00\xc0\xc0\xc0\x00'+\
    '\x80\x80\x80\x00\x00\x00\xff\x00\x00\xff\x00\x00\x00\xff\xff\x00\xff\x00\x00\x00'+\
    '\xff\x00\xff\x00\xff\xff\x00\x00\xff\xff\xff\x00'
    j = 0
    size = 0
    hex = ord(data[15+6])*ord(data[15+8])/8 ## /* size of transparency mask */
    while	(j + 19 + 15 < len(data)):
        nd = ord(data[15+18+j])
        if nd&128 == 128:
            for i in range(nd-128):
                if(size > hex - 1):
                    icodata+=data[15+19+j]
                else:
                    palette[size] = data[15+19+j]
                size+=1
            j = j + 2;
        else:
            for i in range(nd):
                if (size > hex - 1):
                    icodata+= data[15+19+j+i]
                else:
                    palette[size] = data[15+19+j+i]
                size+=1
            j = j + nd + 1;
    for i in range(hex):
        icodata+=palette[i]
    
    page.icon = icodata
##    print 'Icon extracted. Size: ',len(data[19:]),len(icodata)

    
def ForeignType(doc,data,page):
    [type] = struct.unpack('<h', data[0x37:0x39])
    [subtype] = struct.unpack('<h', data[0x39:0x3b])
    [format] = struct.unpack('<L', data[0x44:0x48])
    page.shapes[page.curshape].Foreign.Type = type
    page.shapes[page.curshape].Foreign.Subype = subtype
    page.shapes[page.curshape].Foreign.Format = format
##    page.shapes[page.curshape].Foreign.Sig = sig
##    print 'ForeignType found. Type: %x, Format: %x, Subtype: %x'%(type,format,subtype)
## Type 0 -- stencil
## Type 1
## 0 -- bmp
## 1 -- jpg
## 2 -- gif
## 3 -- tif
## 4 -- png
## Type 2 -- OLE
## Type 4 -- emf/wmf

def ShapeStencil(doc,data,page):
    id = data[0x27:0x37]
    [num] = struct.unpack('<L',data[0x37:0x3b])
    [ownsh] = struct.unpack('<L',data[0x13:0x17])
    tmp = {}
    j = 2
    k = 0
    for i in range ((len(data)-0x36)/4):
        [sh] = struct.unpack('<L',data[0x37+i*4:0x3b+i*4])
        if sh != 0xffffffff and sh != ownsh:
            tmp[sh] = j
            j+=1
        else:
            if sh == 0xffffffff:
                k+=1
    keys = tmp.keys()
    keys.sort()
    list = [tmp[key] for key in keys]
    page.shapes[page.curshape].Foreign.List = list
    page.shapes[page.curshape].Foreign.stflag = 1

def LayerMem(doc,data,page):
##    print 'LayerMem',ix
    pass
    
