# This program was made as part of the sk1project activity to improve UniConvertor
# See http://www.sk1project.org
#
# The code of vsview based on the results of the ALT Linux vsdump project.
# See www.altlinux.ru and http://freshmeat.net/projects/vsdump
#
# Copyright (C) 2007,	Valek Filippov (frob@df.ru)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 or later of the GNU General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import struct 

## Name, Offset, Type, (optional) Dict
## Types:
## 0..7 -- flags
## 8 -- byte
## 9 -- 8 bytes
## 25 -- 2 bytes LE
## 26 -- 4 bytes LE
## 30 -- RGB
##
## Dicts: 
## 0 -- colors
## 1 -- aligns
    
chunks = {
    0x83:(('ID:      ',19,8),('Type:    ',23,8)),
    0x86:(('FG:      ',19,8),('BG:      ',25,30),                        ##Fill^
            ('FillPtrn:',29,8),('ShdwFG:  ',31,8),
            ('ShdwBG:  ',35,30),('ShdwPtrn:',40,8)),
    0x87:(('LeftMrgn:',20,9),('RightMrgn:',29,9),             ##TextBlock
            ('TopMrgn: ',38,9),('BottoMrgn:',47,9),
            ('VAlign:  ',55,8),('TxtBG:   ',56,30),
            ('DfltTab: ',62,9),('TxtDir:  ',82,8)),
    0x89:(('NoFill:  ',19,0),('NoLine:  ',19,0),('NoShow:  ',19,0),('NoSnap:  ',19,3)), ## Geometry
    0x8a:(('X:\t\t',20,9),('Y:\t\t',29,9)),                         ##MoveTo
    0x8b:(('X:\t\t',20,9),('Y:\t\t',29,9)),                         ##LineTo
    0x8c:(('X:\t\t',20,9),('Y:\t\t',29,9),('A:\t\t',38,9)), ##ArcTo
    0x8f:(('X:\t\t',20,9),('Y:\t\t',29,9),('A:\t\t',38,9),
            ('B:\t\t',47,9),('C:\t\t',56,9),('D:\t\t',65,9)), ##Ellipse^
    0x90:(('X:\t\t',20,9),('Y:\t\t',29,9),('A:\t\t',38,9),
            ('B:\t\t',47,9),('C:\t\t',56,9),('D:\t\t',65,9)), ##EllipticalArcTo^
            
    0x9b:(('PinX:\t',20,9),('PinY:\t',29,9),                ##XForm
            ('Width:\t',38,9),('Height:\t',47,9),
            ('LocPinX:',56,9),('LocPinY:',65,9),
            ('Angle:\t',74,9),('FlipX:\t',82,8),('FlipY:\t',83,8),('ResizeMode:',84,8)),
    0x9c:(('PinX:\t',20,9),('PinY:\t',29,9),                ##TxtXForm^
            ('Width:\t',38,9),('Height:\t',47,9),
            ('LocPinX:',56,9),('LocPinY:',65,9),
            ('Angle:\t',74,9)),
    0x9d:(('BeginX:  ',20,9),('BeginY:  ',29,9),            ##XForm1D^
            ('EndX:    ',38,9),('EndY:    ',47,9))
    }


def parse(data):
    type = ord(data[0])
    str = ''
    if chunks.has_key(type):
        args = chunks[type]
        for i in range(len(args)):
            if args[i][2]<8:
                value = ord(data[args[i][1]])&(ord(data[args[i][2]])+1)
            if args[i][2]==8:
                value = ord(data[args[i][1]])
            if args[i][2]==9:
                [value] = struct.unpack('<d',data[args[i][1]:args[i][1]+8])
            if args[i][2]==30:
                value = 'RGB: %02x%02x%02x'%(ord(data[args[i][1]]),ord(data[args[i][1]+1]),ord(data[args[i][1]+2]))
            str=str+args[i][0]+'\t%s\n'%value
    return str
