# This program was made as part of the sk1project activity to improve UniConvertor
# See http://www.sk1project.org
#
# The code of vsview based on the results of the ALT Linux vsdump project.
# See www.altlinux.ru and http://freshmeat.net/projects/vsdump
#
# Copyright (C) 2007,	Valek Filippov (frob@df.ru)
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of version 3 or later of the GNU General Public
# License as published by the Free Software Foundation.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import vsdoc
import gsf
import page
import stencil
import chunks
import inflate
import struct

def vsdopen(fname,doc):
    src = gsf.InputStdio(fname)
    infile = gsf.InfileMSOle(src)
    infchild = infile.child_by_name('VisioDocument')
    ret = 0
    if infchild != None:
        ver_offset = 0x1a
        size_offset = 0x1c
        trlr_offset = 0x24
        pntr = vsdoc.pointer()
        pntr.length = infchild.size()
        itername = 'VisioDocument'
        iter = doc.model.insert(None, 0)
        folder = { "name": itername, "pointer": pntr }
        doc.model.set_value(iter, 0, folder)
        doc.model.set_value(iter, 1, folder["name"])
        pntr.data = infchild.read(pntr.length)
        pntr.path = iter
        doc.version = ord(pntr.data[ver_offset])
        print 'VisioDoc was opened. Version: %x. Length is %x'%(doc.version,pntr.length)
        tr_pntr = vsdoc.pointer()
        if doc.version < 6:
            [tr_pntr.type] = struct.unpack('<h', pntr.data[trlr_offset:trlr_offset+2])
            [tr_pntr.format] = struct.unpack('<h', pntr.data[trlr_offset+2:trlr_offset+4])
            [tr_pntr.address] = struct.unpack('<L', pntr.data[trlr_offset+4:trlr_offset+8])
            [tr_pntr.offset] = struct.unpack('<L', pntr.data[trlr_offset+8:trlr_offset+12])
            [tr_pntr.length] = struct.unpack('<L', pntr.data[trlr_offset+12:trlr_offset+16])
        else:
            [tr_pntr.type] = struct.unpack('<L', pntr.data[trlr_offset:trlr_offset+4])
            [tr_pntr.address] = struct.unpack('<L', pntr.data[trlr_offset+4:trlr_offset+8])
            [tr_pntr.offset] = struct.unpack('<L', pntr.data[trlr_offset+8:trlr_offset+12])
            [tr_pntr.length] = struct.unpack('<L', pntr.data[trlr_offset+12:trlr_offset+16])
            [tr_pntr.format] = struct.unpack('<h', pntr.data[trlr_offset+16:trlr_offset+18])
        iter = doc.model.append(pntr.path, (tr_pntr,'Trailer')) # add trailer
        folder = { "name": 'Trailer', "pointer": tr_pntr }
        doc.model.set_value(iter, 0, folder)
        doc.model.set_value(iter, 1, folder["name"])
        tr_pntr.path = iter
        doc.trailer=iter
        itername = 'Trailer. O/L: %x %x'%(tr_pntr.offset,tr_pntr.length)
        if tr_pntr.format&2 == 2 : #compressed
            res = inflate.inflate(tr_pntr, pntr.data)
            shift = 4
        else:
            res = pntr.data[tr_pntr.offset:tr_pntr.offset+tr_pntr.length]
            shift = 0
        tr_pntr.data = res
        doc.model.foreach(ptr_extract,doc)
        doc.model.foreach(stream_search,doc)
        if len(doc.pages)==0:
            page = vsdoc.page()
            doc.pages.append(page)
            doc.names[0] = 'Test'
            print 'Page appended'
        ret = 1
        
    for i in range(infile.num_children()):
        infchild = infile.child_by_index(i)
        infname = infile.name_by_index(i)
        if infname != 'VisioDocument':
            if ord(infname[0]) == 5: 
                infname = infname[1:]
            pntr = vsdoc.pointer()
            pntr.length = infchild.size()
            pntr.path = None
            itername = infname + ' \t(Size: %x)'%pntr.length
            iter = doc.model.insert(None, 0)
            folder = { "name": itername, "pointer": pntr }
            doc.model.set_value(iter, 0, folder)
            doc.model.set_value(iter, 1, folder["name"])
            pntr.data = infchild.read(pntr.length)
            pntr.path = iter
            if infname == 'SummaryInformation':
                pos = pntr.data.find('EMF')
                if pos !=-1:
                    doc.preview = pntr.data[pos-0x29:]
##                    print 'Position of EMF: %x'%pos
    return ret

streamtype = {0:'Empty pointer',\
    0x14:'Trailer', 0x15:'Page', 0x16:'Colors',0x18:'Fonts',0x1a:'Styles',0x1d:'Stencils', 0x1e:'Stencil Page',\
    0x23:'Icon',0x27:'Pages',0x29:'Windows',0x2a:'Window',0x2e:'EventList',0x2f:'EventItem',\
    0x31:'Document',0x32:'NameList',0x33:'Name',\
    0xd7:'FontFace',0xd8:'FontFaces'}

def stream_search(model,path,iter,doc):
    pntr = model.get_value(iter,0)["pointer"]
    if pntr.type == 0x32:
        getnames(doc,iter)
    if pntr.type == 0x16:
        getcolors(doc,iter)
    if pntr.type == 0x18 or pntr.type == 0xd8:
        getfonts(doc,iter)
    if pntr.type == 0x1d:
        stencil.search(doc,iter)
    if pntr.type == 0x27:
        doc.psiter = iter
        page.search(doc,iter)

    if pntr.format>>4 == 0xd and pntr.type != 0x18 and pntr.type!=0x15 and pntr.type!=0x46 and pntr.type!=0x47 and pntr.type!=0x48:
        chunks.chunk_parse(doc,pntr)
        
     
def ptr_search(model, pntr,shift,doc):
    res = pntr.data
    parent = pntr.path
    if doc.version<6:
        plength=16
        [num] = struct.unpack('<h', res[0xa+shift:0xc+shift])
        offset = 12+shift
        if pntr.type == 0x14:
            [num] = struct.unpack('<h', res[0x82+shift:0x84+shift])
            offset = 0x84+shift
        if pntr.type == 0x1d:
            [num] = struct.unpack('<h', res[0x1e+shift:0x20+shift])
            offset = 0x20+shift
        if pntr.type == 0x1e:
            [num] = struct.unpack('<h', res[0x36+shift:0x38+shift])
            offset = 0x38+shift
        if pntr.type == 0x4e:
            [num] = struct.unpack('<h', res[0x1e+shift:0x20+shift])
            offset = 0x20+shift

    else:
        [offset] = struct.unpack('<L', res[shift:shift+4])
        [num] =  struct.unpack('<L', res[offset+shift:offset+shift+4])
        offset = offset+8+shift
        plength=18
    for i in range(num):
            pntr = vsdoc.pointer()
            if doc.version < 6:
                [pntr.type] = struct.unpack('<h', res[offset+i*plength:offset+i*plength+2])
                [pntr.format] = struct.unpack('<h', res[offset+i*plength+2:offset+i*plength+4])
                [pntr.address] = struct.unpack('<L', res[offset+i*plength+4:offset+i*plength+8])
                [pntr.offset] = struct.unpack('<L', res[offset+i*plength+8:offset+i*plength+12])
                [pntr.length] = struct.unpack('<L', res[offset+i*plength+12:offset+i*plength+16])
            else:
                [pntr.type] = struct.unpack('<L', res[offset+i*plength:offset+i*plength+4])
                [pntr.address] = struct.unpack('<L', res[offset+i*plength+4:offset+i*plength+8])
                [pntr.offset] = struct.unpack('<L', res[offset+i*plength+8:offset+i*plength+12])
                [pntr.length] = struct.unpack('<L', res[offset+i*plength+12:offset+i*plength+16])
                [pntr.format] = struct.unpack('<h', res[offset+i*plength+16:offset+i*plength+18])
            pntr.path = parent
            itername = '%02x \t\tA/O/L/F: \t%07x \t%04x \t%04x \t%02x'%(pntr.type,pntr.address,pntr.offset,pntr.length,pntr.format)
##            print 'T/A/O/L/F: %02x \t%07x \t%04x \t%04x \t%02x'%(pntr.type,pntr.address,pntr.offset,pntr.length,pntr.format)
            if streamtype.has_key(pntr.type):
                itername = streamtype[pntr.type]+' \tA/O/L/F: \t%07x \t%04x \t%04x \t%02x'%(pntr.address,pntr.offset,pntr.length,pntr.format)
            else:
                if chunks.chunktype.has_key(pntr.type):
                    itername = chunks.chunktype[pntr.type]+' \tA/O/L/F: \t%07x \t%04x \t%04x \t%02x'%(pntr.address,pntr.offset,pntr.length,pntr.format)
            if pntr.type == 0:
                itername = '(Empty pointer)'
            iter = model.append(pntr.path,(pntr,itername))
            folder = { "name": itername, "pointer": pntr }
            model.set_value(iter, 0, folder)
            model.set_value(iter, 1, folder["name"])
            pntr.path = iter

def ptr_extract(model,path,iter,doc):
    pntr = model.get_value(iter,0)["pointer"]
    vsdpntr = model.get_value(model.get_iter_first(),0)["pointer"]
    if pntr.format == 0:
        return
    vsd = vsdpntr.data
    if pntr.format&2 == 2 : #compressed
        res = inflate.inflate(pntr, vsd)
        shift = 4
    else:
        res = vsd[pntr.offset:pntr.offset+pntr.length]
        shift = 0
    pntr.data = res
    if pntr.format>>4 == 5 and pntr.type != 0x16:
        ptr_search(model,pntr,shift,doc)
    

def getcolors(doc,citer):
##    print 'Color entries were found.'
    colors = doc.colors
    pntr = doc.model.get_value(citer,0)["pointer"]
    res = pntr.data
    clrnum = ord(res[6])
    for i in range(clrnum):
        clr = vsdoc.color()
        clr.r = ord(res[8+i*4])
        clr.g = ord(res[9+i*4])
        clr.b = ord(res[10+i*4])
        clr.a = 0
        colors.append(clr)
## small addition for Colors
        cpntr = vsdoc.pointer()
        iter = doc.model.append(pntr.path,(cpntr,'Colors'))
        itername = 'Color: %02x'%i+'\tRGB: #%02x%02x%02x'%(clr.r,clr.g,clr.b) 
        folder = { "name": itername, "pointer": cpntr }
        doc.model.set_value(iter, 0, folder)
        doc.model.set_value(iter, 1, folder["name"])
        cpntr.path = iter
        cpntr.data = res[8+i*4:11+i*4]
        
    for i in range(len(colors)):
        clr = colors[i]
##        print ' IX=%u  RGB=#%02x%02x%02x'%(i, clr.r,clr.g,clr.b)

def getnames(doc, niter):
##    print 'Names were found...',doc.model.iter_n_children(niter)
    names = doc.names
    for i in range(doc.model.iter_n_children(niter)):
            nameiter = doc.model.iter_nth_child(niter, i)
            npntr = doc.model.get_value(nameiter,0)["pointer"]
            nitername = doc.model.get_value(nameiter,0)["name"]
            shift = 4
            if npntr.format&2 == 2 : #compressed
                shift = 8
            if doc.version == 6:
                name = unicode(npntr.data[shift:],'cp1251').encode('utf-8')  ## FIXME! have to check locale in the file
            if doc.version == 11:
                name = unicode(npntr.data[shift:],'utf-16').encode('utf-8')
            cntr = name.find('\00')
            names[i] = name[0:cntr]
##            print 'Name: ',i,name
            nitername = nitername + '   \t'+name
            doc.model.set_value(nameiter,1,nitername)

def getfonts(doc,iter):
##    print 'Font entries were found...'
    fonts = doc.fonts
    pntr = doc.model.get_value(iter,0)["pointer"]
    # for Visio 2k/2k2 this pntr.data is type D stream (made of chunks)
    # for Visio 2k3 this pntr->children are FontEntries
    if doc.version == 6:
        chunks.chunk_parse(doc,pntr)
        fiter = doc.model.iter_nth_child(iter, 0)
        for i in range(doc.model.iter_n_children(fiter)):
            fntiter = doc.model.iter_nth_child(fiter, i)
            fpntr = doc.model.get_value(fntiter,0)["pointer"]
            fitername = doc.model.get_value(fntiter,0)["name"]
            fname = fpntr.data[25:len(fpntr.data)]
            [fn] = struct.unpack('<h',fpntr.data[4:6])
            cntr = fname.find('\00')
            fonts[fn] = fname[0:cntr]
            fitername = fitername + '   \t'+fname
            doc.model.set_value(fntiter,1,fitername)

    if doc.version == 11: ## check for font substitution
        num = 0
        fiter = doc.model.iter_nth_child(iter, 0)
        for i in range(doc.model.iter_n_children(fiter)):
            fntiter = doc.model.iter_nth_child(fiter, i)
            fpntr = doc.model.get_value(fntiter,0)["pointer"]
            fname = unicode(fpntr.data[8:72],'utf-16').encode('utf-8')
            cntr = fname.find('\00')
            fonts[i] = fname[0:cntr]
            fitername = fitername + '   \t'+fname
            doc.model.set_value(fntiter,1,fitername)
            print 'Font ID: %u'%(i),' Font name: ',fname[0:cntr]

        for i in range(doc.model.iter_n_children(iter)):
            child = doc.model.iter_nth_child(iter,i)        
            childptr = doc.model.get_value(child,0)["pointer"]
            if childptr.type == 0x3f:
            # looking for font idx substitution
                fidx = doc.model.iter_nth_child(doc.model.iter_nth_child(child,0),1)
                fidxpntr = doc.model.get_value(fidx,0)["pointer"]
                print 'Found fidxptr! T/A/O/L/F: %02x \t%07x \t%04x \t%04x \t%02x'%(fidxpntr.type,fidxpntr.address,fidxpntr.offset,fidxpntr.length,fidxpntr.format)
                print 'Data length: ',len(fidxpntr.data)
                length = len(fidxpntr.data)
                if length > 10:
                    [f]= struct.unpack('<L',fidxpntr.data[10:14])
                    if f > 4:
                        shift = 10
                    else:
                        shift = 14
                    for i in range((length-shift)/4):
                        [fn] = struct.unpack('<L',fidxpntr.data[shift+i*4:shift+i*4+4])
##                        print 'Substitution (fn/length/i): ', fn,length,i
                        doc.fonts[fn] = doc.fonts[length+i]



